/******************************************************************************
*
*      FILE: qs_sl_io.c
*   VERSION: 4.00
*      DATE: May 2009
* COPYRIGHT: QAS Ltd
*
* C sample code for integrating the QAS Pro Single Line search engine (V4.x).
*
* This module contains all functions that are responsible for the input and
* output of information (i.e. interacting with the user). It should be
* modifed according to the requirements of the host application into which the
* address capture process is being integrated.
*
* The functions use simple console based I/O (using the standard C functions
* printf and gets).
*
* For simplicity, all picklists are displayed without any consideration for the
* number of items that can be viewed (i.e. large picklists will scroll off the
* top of the screen). Integrations that are similarly limited should break the
* picklist up into multiple pages (screens) with appropriate shortcuts for
* navigating between them.
*
* See qs_sl.c for more information.
*
******************************************************************************/


#include "qs_sl.h"




/**************
* QS_ReportError ()
*
* Reports an error to the user.
*
* Input:
*   ptSearch    : Search context
*   iError      : QAS Pro API error code
*   psMessage   : Additional message to display
*
* Return:
*   None.
***************/

void QS_ReportError (Search_t *ptSearch, int iError, char *psMessage)
{
    char    sError [ERROR_LENGTH];          /* Buffer for obtaining QAS error message */

    /* Display error message */
    printf ("\nError: %s\n", psMessage);

    /* Obtain additional error message from QAS Pro */
    QA_ErrorMessage (iError, sError, sizeof (sError));

    printf ("QAS error: %s\n\n", sError);
}
    


/**************
* QS_DisplayMessage()
*
* Display a non-serious message to the user. Typically, such messages are
* displayed in a status bar (often along with some form of additional
* highlighting).
*
* Input:
*   psMessage  : Message text
*
* Return:
*   None
***************/

void QS_DisplayMessage (char *psMessage)
{
    printf("\n%s\n", psMessage);
}


/**************
* QS_GetSearchParameters()
*
* Request data set and search details. This represents the main dialog used to
* select the country data set and enter the single line search text.
*
* Input:
*   ptSearch   : Search context
*   psDataSet  : Current data set
*
* Output:
*   psDataSet  : Selected data set
*   psSearch   : Search to be performed
*
* Return:
*   0 = Search cancelled / error occurred
*   1 = Search parameters entered
***************/

void QS_GetLayoutParameter (char *psLayout)
{
    char  sInput[SEARCH_LENGTH];        /* Input buffer */
    char *psNewLine = NULL;             /* Newline character */


    /* Prompt the user if they wish to specify a layout */
    /* If they do not, it will default to the "< Default >" layout */
    /* that is defined in the "QADefault" section in QAWORLD.INI */
    printf ("\nLayout [Enter='%s']: ", psLayout);


    /* This gets a string entered by the user from the standard input */
    fgets (sInput, SEARCH_LENGTH, stdin);

    /* The 'fgets' function may add a newline character to the buffer */
    /* which has to be stripped out */
    psNewLine = strchr (sInput, '\n');
    if (psNewLine != NULL)
    {
        *psNewLine = '\0';
    }


    /* If a layout was specified, copy it to be used */
    if (sInput[0] != '\0')
    {
        strcpy (psLayout, sInput);
    }
}


/**************
* QS_GetSearchParameters()
*
* Request data set and search details. This represents the main dialog used to
* select the country data set and enter the single line search text.
*
* Input:
*   ptSearch   : Search context
*   psDataSet  : Current data set
*
* Output:
*   psDataSet  : Selected data set
*   psSearch   : Search to be performed
*
* Return:
*   0 = Search cancelled / error occurred
*   1 = Search parameters entered
***************/

int QS_GetSearchParameters (Search_t *ptSearch, char *psDataSet, char *psSearch)
{
    char  sInput [SEARCH_LENGTH];           /* Input buffer */
    char *psNewLine;                        /* New line character */

    /* Loop until a data set has been selected. The user can */
    /* also press enter to select the current default, ? to  */
    /* view a list of all available data sets, or q to quit. */

    do
    {
        printf ("\nCountry [Enter=%s,?=list,q=quit]: ", psDataSet);

        /* This gets a string entered by the user from the standard input */
        fgets (sInput, SEARCH_LENGTH, stdin);

        /* The 'fgets' function may add a newline character to the buffer */
        /* which has to be stripped out */
        psNewLine = strchr (sInput, '\n');
        if (psNewLine != NULL)
        {
            *psNewLine = '\0';
        }

        if (*sInput == 'Q' || *sInput == 'q')
        {
            return (0);
        }
        else if (*sInput == '?')
        {
            QS_ViewDataSets (ptSearch);
        }
        else if (*sInput != '\0')
        {
            strcpy (psDataSet, sInput);
        }
    }
    while (*sInput == '?');


    /* Get the single line search text */

    do
    {
        /* The user can type an exclaimation mark at this point to */
        /* display the system information text. */
        printf ("\nSearch [!=sysinfo]: ");


        /* This gets a string entered by the user from the standard input */
        fgets (psSearch, SEARCH_LENGTH, stdin);

        /* The 'fgets' function may add a newline character to the buffer */
        /* which has to be stripped out */
        psNewLine = strchr (psSearch, '\n');
        if (psNewLine != NULL)
        {
            *psNewLine = '\0';
        }

        /* Display the system information if an exclaimation mark was entered */
        if (*psSearch == '!')
        {
            QS_ViewSystemInfo (ptSearch);
        }
    }
    while (*psSearch == '!');

    return (1);
}




/**************
* QS_ViewDataSets()
*
* Display list of available data sets.
*
* Input:
*   ptSearch : Search context
*
* Return:
*   None
***************/

void QS_ViewDataSets (Search_t *ptSearch)
{
    int     iError = 0;                 /* QAS Pro API error */
    int     iCount = 0;                 /* Count of data sets */
    int     iIndex;                     /* Used to iterate through data sets */
    char    sIso  [DATASET_LENGTH];     /* Data set information */
    char    sName [DATANAME_LENGTH];    /* Data name information */


    /* Obtain count of data sets */
    iError = QA_GetDataCount (ptSearch->iHandle, &iCount);


    /* Loop through each data set and display it's ISO */
    for (iIndex = 0; iError == 0 && iIndex < iCount; iIndex ++)
    {
        iError = QA_GetData (ptSearch->iHandle, iIndex, sIso, sizeof (sIso), sName, sizeof (sName));

        if (iError == 0)
        {
            printf("%s - %s\n", sIso, sName);
        }
    }


    /* Report any errors encountered */
    if (iError != 0)
    {
        QS_ReportError (ptSearch, iError, "Viewing data sets");
    }
}



/**************
* QS_ViewSystemInfo()
*
* Output system information.
*
* Input:
*   ptSearch : Search context
*
* Return:
*   None
***************/

void QS_ViewSystemInfo (Search_t *ptSearch)
{
    int     iError  = 0;                    /* QAS Pro API error */
    int     iCount  = 0;                    /* Count of system information lines */
    int     iLineNo;                        /* Line number */
    char    sLine [SEARCH_LENGTH];          /* System information text */


    printf ("\nSystem Information:\n\n");


    /* Obtain count of information lines */
    iError = QA_GenerateSystemInfo (ptSearch->iHandle,
                                    qasysinfo_SYSTEM,
                                    &iCount);


    /* Loop through each line, displaying to the user */
    for (iLineNo = 0; iError == 0 && iLineNo < iCount; iLineNo++)
    {
        iError = QA_GetSystemInfo (ptSearch->iHandle,
                                   iLineNo,
                                   sLine,
                                   sizeof (sLine));

        if (iError == 0)
        {
            printf("  %s\n", sLine);
        }
    }


    /* Report any errors encountered */
    if (iError != 0)
    {
        QS_ReportError (ptSearch, iError, "Displaying system information");
    }
}



/**************
* QS_NavigatePickLists()
*
* Perform all picklist navigation. This function is responsible for displaying
* picklists and interacting with the user. It will only return once the user
* has selected a final (leaf-node) address.
*
* Input:
*   ptSearch : Search context
*
* Return:
*   0 = Search cancelled
*   1 = Leaf-node address selected
***************/

int QS_NavigatePickLists (Search_t *ptSearch)
{
    int     iError          = 0;                /* QAS Pro error code */
    int     iAvailable      = 0;                /* Picklist size */
    int     iPotential      = 0;                /* Potential matches */
    long    lFlags          = 0;                /* Picklist properties */
    long    lResultFlags    = 0;                /* Result properties */
    char    sRefine [SEARCH_LENGTH];            /* Refinement text */
    char    sSearch [SEARCH_LENGTH];            /* Text to pass to search */
    char    sInput  [SEARCH_LENGTH];            /* Input buffer */
    char    sPrompt [SEARCH_LENGTH];            /* Picklist prompt */
    char    sLine   [SEARCH_LENGTH];            /* Result line */
    char   *psNewLine;                          /* New line character */


    /* Reset the refinement text */
    sRefine[0] = '\0';


    /* Loop whilst we have a valid (active) picklist and */
    /* the user has not selected a leaf-node address and */
    /* there are no problems retrieving and displaying   */
    /* the picklist of results. */

    while (iError == 0
           &&   ptSearch->iLevel >= 0
           &&   ptSearch->iAddress < 0)
    {

        /* Loop through the process of obtaining search status and  */
        /* automatically stepping in to the picklist as advised by  */
        /* the API. */

        /* Obtain the status of the search */
        iError = QA_GetSearchStatus (ptSearch->iHandle, &iAvailable, &iPotential, &lFlags);

        do
        {
            /* Perform automatic step in if appropiate. The search status returns */
            /* flags when it deems the current picklist suitable to automatically */
            /* step into without consulting the user. */
            if (iError == 0)
            {
                /* The flag qastate_AUTOSTEPINSAFE is returned when the current picklist */
                /* has only one result and it is safe to automatically step into the first item. */

                /* The flag qastate_AUTOSTEPINPASTCLOSE is returned when the current picklist */
                /* has only one exact result, but more lesser results. It is advisable to step. */
                /* into the first item */

                /* The flag qastate_AUTOFORMATSAFE is returned when the current picklist */
                /* has only one result and it is safe to automatically format the first item. */
                if (lFlags & (qastate_AUTOSTEPINSAFE | qastate_AUTOSTEPINPASTCLOSE | qastate_AUTOFORMATSAFE))
                {
                    iError = QS_SelectPickListItem (ptSearch, iAvailable, 0);
                }

                /* For this integration, we also wich to automatically select a picklist item */
                /* if we have just refined and generated a single picklist result */
                else if (iError == 0
                         &&  iAvailable == 1
                         &&  sRefine[0] != '\0')
                {
                    iError = QA_GetResult (ptSearch->iHandle, 0, sLine, sizeof (sLine), NULL, &lResultFlags);

                    /* We wish to automatically select if it is a full address or can be */
                    /* stepped into, unless it is an informational item */
                    if (iError == 0
                        &&  lResultFlags & (qaresult_FULLADDRESS | qaresult_CANSTEP)
                        &&  (lResultFlags & qaresult_INFORMATION) == 0)
                    {
                        iError = QS_SelectPickListItem (ptSearch, iAvailable, 0);

                        if (iError == 0)
                        {
                            sRefine[0] = '\0';
                        }
                    }
                }

                if (iError == 0)
                {
                    /* Refresh the status of the search */
                    iError = QA_GetSearchStatus (ptSearch->iHandle, &iAvailable, &iPotential, &lFlags);
                }
            }
        }
        while (iError == 0
               &&   (lFlags & (qastate_AUTOSTEPINSAFE | qastate_AUTOSTEPINPASTCLOSE | qastate_AUTOFORMATSAFE))
               &&   ptSearch->iAddress < 0);



        /* Check that we did not auto step and format */
        /* If we did, then we have finished and can break from the loop */
        if (iError == 0 && ptSearch->iAddress >= 0)
        {
            break;
        }

        /* Display picklist */
        if (iError == 0)
        {
            iError = QS_UpdatePickList (ptSearch, iAvailable, iPotential);
        }

        /* Obtain picklist prompt */
        if (iError == 0)
        {
            iError = QA_GetPrompt (ptSearch->iHandle, 0, sPrompt, sizeof(sPrompt), NULL, NULL, 0);
        }


        /* Display prompt to user */
        if (iError == 0)
        {
            printf ("\n%s [#n or text]: %s", sPrompt, sRefine);


            /* Obtain input from user */
            fgets (sInput, SEARCH_LENGTH, stdin);

            /* The 'fgets' function may add a newline character to the buffer */
            /* which has to be stripped out */
            psNewLine = strchr (sInput, '\n');
            if (psNewLine != NULL)
            {
                *psNewLine = '\0';
            }


            if (*sInput == '#')
            {
                /* Attempt to step into picklist item #n */

                iError = QS_SelectPickListItem (ptSearch, iAvailable, atoi(sInput + 1) - 1);

                if (iError == 0)
                {
                    *sRefine = '\0';
                }
            }
            else if (*sInput != '\0')
            {
                /* User has specified additional refinement text */

                strcat (sRefine, sInput);


                /* Append a space to the refinement text so to use  */
                /* whole number matching. Refining upon "1 " for    */
                /* example will not match to "10", giving shorter   */
                /* picklists suitable for console-based input       */
                strcpy (sSearch, sRefine);
                strcat (sSearch, " ");

                iError = QA_Search (ptSearch->iHandle, sSearch);
            }
            else
            {
                /* User has requested to step back (cancel the last    */
                /* action). If the user had previously refined the     */
                /* picklist then clear this otherwise step back to the */
                /* previous picklist level. If the user is already at  */
                /* level 0 then the search will be cancelled.          */

                if (*sRefine != '\0')
                {
                    iError = QA_Search (ptSearch->iHandle, "");

                    if (iError == 0)
                    {
                        *sRefine = '\0';
                    }
                }
                else if (ptSearch->iLevel == 0)
                {
                    /* Cancel the search */
                    ptSearch->iLevel = -1;
                }
                else
                {
                    iError = QA_StepOut (ptSearch->iHandle);

                    if (iError == 0)
                    {
                        /* Decrement search level */
                        ptSearch->iLevel --;
                    }
                }
            }
        }
    }

    /* Report any error while navigating picklist */
    if (iError != 0)
    {
        QS_ReportError (ptSearch, iError, "Error while navigating, search cancelled");

        return 0;
    }

    /* Determine if the user has selected a leaf-node address */

    return (ptSearch->iAddress >= 0);
}



/**************
* QS_StartPickListDisplay()
*
* Display the picklist header. This generally consists of the search history
* (providing the user with "context" for the displayed picklist) and the total
* number of matches available.
*
* Input:
*   ptSearch   : Search context
*   iPotential : Potential number of matches to report
*
* Return:
*   None
***************/

void QS_StartPickListDisplay (Search_t *ptSearch, int iPotential)
{
    printf ("\nLevel %d: \"%s\" - %d matches\n\n",
            ptSearch->iLevel,
            ptSearch->asHistory[ptSearch->iLevel],
            iPotential);
}


/**************
* QS_DisplayPickListItem()
*
* Displays the given picklist item.
*
* Input:
*   psAddress  : Picklist item to display
*
* Return:
*   None
***************/

void QS_DisplayPickListItem (char *psAddress)
{
    printf ("%s\n", psAddress);
}



/**************
* QS_DisplayAddress()
*
* Display the formatted leaf-node address.
*
* Input:
*   ptSearch : Search context
*   iCount   : Count of address lines
*
* Return:
*   0       : Success
*   != 0    : Failure
***************/

int QS_DisplayAddress (Search_t *ptSearch, int iCount)
{
    int     iError  = 0;                /* QAS Pro error code */
    int     iIndex  = 0;                /* Used to iterate through address lines */
    char    sLabel [SEARCH_LENGTH];     /* Formatted address information */
    char    sLine  [SEARCH_LENGTH];     /* Formatted address */


    printf("\n");

    /* Loop through each line in turn */
    for (iIndex = 0; iError == 0 && iIndex < iCount; iIndex++)
    {
        iError = QA_GetFormattedLine (ptSearch->iHandle, iIndex, sLine, sizeof (sLine), sLabel, sizeof (sLabel), NULL);

        if (iError == 0)
        {
            printf ("%12s: %s\n", sLabel, sLine);
        }
    }

    return iError;
}



/******************************************************************************
******************* End of file qs_sl_io.c ************************************
******************************************************************************/
